/**
  **************************************************************************
  * @file     hid_iap_user.c
  * @version  v2.1.1
  * @date     2022-07-22
  * @brief    usb msc iap user file
  **************************************************************************
  *                       Copyright notice & Disclaimer
  *
  * The software Board Support Package (BSP) that is made available to
  * download from Artery official website is the copyrighted work of Artery.
  * Artery authorizes customers to use, copy, and distribute the BSP
  * software and its related documentation for the purpose of design and
  * development in conjunction with Artery microcontrollers. Use of the
  * software is governed by this copyright notice and the following disclaimer.
  *
  * THIS SOFTWARE IS PROVIDED ON "AS IS" BASIS WITHOUT WARRANTIES,
  * GUARANTEES OR REPRESENTATIONS OF ANY KIND. ARTERY EXPRESSLY DISCLAIMS,
  * TO THE FULLEST EXTENT PERMITTED BY LAW, ALL EXPRESS, IMPLIED OR
  * STATUTORY OR OTHER WARRANTIES, GUARANTEES OR REPRESENTATIONS,
  * INCLUDING BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
  * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.
  *
  **************************************************************************
  */

#include "msc_iap_user.h"
#include "msc_class.h"
#include "string.h"
#include "ff.h"
#include "at32_spiflash.h"

/** @addtogroup UTILITIES_examples
  * @{
  */

/** @addtogroup USB_IAP_bootloader
  * @{
  */

iap_info_type iap_info;
void (*pftarget)(void);
FATFS fs;
FIL file;
BYTE work[FF_MAX_SS];
#define PDRV_STR    "2:"
#define FILENAME    "2:/app.bin"
#define RENAME_SUCCESS_FILE  "2:/success.bin"
#define RENAME_ERROR_FILE  "2:/error.bin"


uint32_t flash_write_data(uint32_t address, uint8_t *data, uint32_t len);
error_status crc_cal(uint32_t addr, uint32_t n_byte, uint32_t file_crc);
uint32_t buffer_crc(uint8_t *buffer, uint32_t n_byte);
error_status update_upgrade_flag(error_status status);
void iap_set_upgrade_flag(void);
void iap_clear_upgrade_flag(void);

/**
  * @brief  check upgrade bin
  * @param  none
  * @retval status
  */
error_status iap_check_upgrade(void)
{
  FRESULT ret;
  error_status status = SUCCESS;

  ret = f_mount(&fs, PDRV_STR, 1);
  if(ret == FR_OK)
  {
    ret = f_open(&file, FILENAME, FA_READ | FA_WRITE);
    if(ret != FR_OK)
    {
      status = ERROR;
    }
  }
  else
  {
    status = ERROR;
  }
  
  if(status == ERROR)
  {
    f_close(&file);
    f_unmount(PDRV_STR);
  }
  
  return status;
}

/**
  * @brief  upgrade process
  * @param  none
  * @retval upgrade status
  */
error_status iap_upgrade_process(void)
{
  FILINFO fno;
  FRESULT ret;
  error_status status = SUCCESS;
  uint32_t file_size = 0, read_sector = 0;
  uint32_t sector_num = 0;
  uint32_t rw_byte = 0;
  uint32_t address = iap_info.app_address, write_size = 0;
  uint32_t file_crc = 0;
  ret = f_stat(FILENAME, &fno);
  if(ret != FR_OK)
  {
    status = ERROR;
    goto UP_DONE;
  }
  if(fno.fsize <= 0 || fno.fsize > iap_info.flash_end_address - iap_info.app_address)
  {
    status = ERROR;
    goto UP_DONE;
  }
  
  file_size = fno.fsize;
  
  sector_num = file_size / flash_spi_sectorsize();
  if(file_size % flash_spi_sectorsize())
  {
    sector_num += 1;
  }
  
  crm_periph_clock_enable(CRM_CRC_PERIPH_CLOCK, TRUE);
  crc_data_reset();
  
  /* start upgrade */
  iap_clear_upgrade_flag();
  for(read_sector = 0; read_sector < sector_num; read_sector ++)
  {
    ret = f_read(&file,  iap_info.iap_fifo, flash_spi_sectorsize(), &rw_byte);
    if(ret != FR_OK)
    {
      status = ERROR;
      goto UP_DONE;
    }
    if(write_size + rw_byte >= file_size)
    {
      flash_write_data(address, iap_info.iap_fifo, file_size - write_size);
      break;
    }
    else
    {
      flash_write_data(address, iap_info.iap_fifo, rw_byte);
      buffer_crc(iap_info.iap_fifo, rw_byte);
    }
    write_size += rw_byte;
    address += rw_byte;
  }
  
  file_crc = buffer_crc(iap_info.iap_fifo, file_size - write_size);
  
  /* crc check */
  if(crc_cal(iap_info.app_address, file_size, file_crc) != SUCCESS)
  {
    status = ERROR;
  }
UP_DONE:
  f_close(&file);
  f_unlink(RENAME_SUCCESS_FILE);
  f_unlink(RENAME_ERROR_FILE);
  if(status == SUCCESS)
  {
    ret = f_rename (FILENAME, RENAME_SUCCESS_FILE);
    iap_set_upgrade_flag();
  }
  else
  {
    f_rename (FILENAME, RENAME_ERROR_FILE);
  }
  
  f_unmount(PDRV_STR);
  return status;
}


/**
  * @brief  write data to flash
  * @param  address: flash address
  * @param  data: pointer to array to store data
  * @param  len: data length
  * @retval write status
  */
uint32_t flash_write_data(uint32_t address, uint8_t *data, uint32_t len)
{
  uint32_t waddr = address;
  uint32_t wlength = len;
  uint32_t i_index = 0;
  if((address + len) > iap_info.flash_end_address )
  {
    return 1;
  }

  if((waddr & iap_info.sector_msk) == 0)
  {
    flash_unlock();
    if(len > iap_info.sector_size)
    {
      for(; wlength >= iap_info.sector_size; wlength -= iap_info.sector_size)
      {
        flash_sector_erase(waddr);
        waddr += iap_info.sector_size;
      }
      if(wlength > 0)
      {
        flash_sector_erase(waddr);
      }
    }
    else
    {
      flash_sector_erase(waddr);
    }

    flash_lock();
  }

  waddr = address;
  flash_unlock();
  for(i_index = 0; i_index < len; i_index ++)
  {
    flash_byte_program(waddr+i_index, data[i_index]);
  }
  flash_lock();

  return 0;
}

/**
  * @brief  crc cal
  * @param  addr: start address
  * @param  n_byte: number of byte
  * @retval crc status
  */
error_status crc_cal(uint32_t addr, uint32_t n_byte, uint32_t file_crc)
{
  uint32_t *paddr = (uint32_t *)addr;
  uint32_t wlen = n_byte / sizeof(uint32_t), remain = 0, remain_msk = 0xFFFFFFFF;
  uint32_t value, i_index = 0, flash_crc = 0;
  crm_periph_clock_enable(CRM_CRC_PERIPH_CLOCK, TRUE);
  crc_data_reset();
  
  /* flash crc */
  if(n_byte % sizeof(uint32_t) != 0)
  {
    remain = n_byte % sizeof(uint32_t);
  }

  for(i_index = 0; i_index < wlen; i_index ++)
  {
    value = *paddr;
    crc_one_word_calculate(value);
    paddr ++;
  }
  if(remain != 0)
  {
    value = *paddr;
    remain_msk <<= 8 * remain;
    value |= remain_msk;
    crc_one_word_calculate(value);
  }
  flash_crc = crc_data_get();
  
  if(flash_crc != file_crc)
  {
    return ERROR;
  }
  return SUCCESS;
}

/**
  * @brief  buffer_crc
  * @param  buffer: data pointer
  * @param  n_byte: number of byte
  * @retval crc value
  */
uint32_t buffer_crc(uint8_t *buffer, uint32_t n_byte)
{
  uint32_t i_index = 0;
  uint32_t wlen = n_byte / sizeof(uint32_t), remain = 0, remain_msk = 0xFFFFFFFF;
  uint32_t *value = (uint32_t *)buffer;
  if(n_byte % sizeof(uint32_t) != 0)
  {
    remain = n_byte % sizeof(uint32_t);
  }
  for(i_index = 0; i_index < wlen; i_index ++)
  {
    crc_one_word_calculate(*value);
    value ++;
  }
  if(remain != 0)
  {
    remain_msk <<= 8 * remain;
    crc_one_word_calculate(*value | remain_msk);
  }
  return crc_data_get();
}

/**
  * @brief  clear iap upgrade flag
  * @param  none
  * @retval none
  */
void iap_clear_upgrade_flag(void)
{
  flash_unlock();
  flash_sector_erase(iap_info.flag_address);
  flash_lock();
}

/**
  * @brief  set iap upgrade complete flag
  * @param  none
  * @retval none
  */
void iap_set_upgrade_flag(void)
{
  flash_unlock();
  flash_word_program(iap_info.flag_address, IAP_UPGRADE_COMPLETE_FLAG);
  flash_lock();
}

/**
  * @brief  get iap upgrade complete flag
  * @param  none
  * @retval the status of the iap flag
  */
error_status iap_get_upgrade_flag(void)
{
  uint32_t flag_address = iap_info.flag_address;
  if(*((uint32_t *)flag_address) == IAP_UPGRADE_COMPLETE_FLAG)
  {
    return SUCCESS;
  }
  else
  {
    return ERROR;
  }
}

/**
  * @brief  jump to app
  * @param  none
  * @retval none
  */
void jump_to_app(uint32_t address)
{
  uint32_t stkptr, jumpaddr;
  stkptr = *(uint32_t *)address;
  jumpaddr = *(uint32_t *)(address + sizeof(uint32_t));

  /* disable nvic irq and periph clock, clear pending */
  nvic_irq_disable(USBFS_L_CAN1_RX0_IRQn);
  __NVIC_ClearPendingIRQ(USBFS_L_CAN1_RX0_IRQn);
  crm_periph_clock_enable(CRM_GPIOA_PERIPH_CLOCK, FALSE);
  crm_periph_clock_enable(CRM_USB_PERIPH_CLOCK, FALSE);

  crm_periph_reset(CRM_USB_PERIPH_RESET, TRUE);
  crm_periph_reset(CRM_USB_PERIPH_RESET, FALSE);

  __set_MSP(stkptr);
  pftarget = (void (*) (void))jumpaddr;
  pftarget();
}

/**
  * @brief  iap init
  * @param  none
  * @retval none
  */
void iap_init(void)
{
  iap_info.flash_size = KB_TO_B(FLASH_SIZE_REG());
  if(iap_info.flash_size  < KB_TO_B(256))
  {
    iap_info.sector_size = SECTOR_SIZE_1K;
    iap_info.sector_msk = FLASH_SECTOR_1K_ALLGNED;
  }
  else
  {
    iap_info.sector_size = SECTOR_SIZE_2K;
    iap_info.sector_msk = FLASH_SECTOR_2K_ALLGNED;
  }

  iap_info.flash_start_address = FLASH_BASE;
  iap_info.flash_end_address = iap_info.flash_start_address + iap_info.flash_size;

  iap_info.app_address = FLASH_APP_ADDRESS;
  iap_info.flag_address = iap_info.app_address -  iap_info.sector_size;

  iap_info.fifo_length = 0;
  iap_info.iap_address = 0;
}

/**
  * @}
  */

/**
  * @}
  */
